@@ -11,7 +11,7 @@ MIME::EncWords - deal with RFC 2047 encoded words (improved)
 
 I<L<MIME::EncWords> is aimed to be another implimentation
 of L<MIME::Words> so that it will achieve more exact conformance with
-RFC 2047 (former RFC 1522) specifications.  Additionally, it contains
+RFC 2047 (formerly RFC 1522) specifications.  Additionally, it contains
 some improvements.
 Following synopsis and descriptions are inherited from its inspirer,
 then added descriptions on improvements (B<**>) or changes and
@@ -130,7 +130,7 @@ if (MIME::Charset::USE_ENCODE) {
 #------------------------------
 
 ### The package version, both in 1.23 style *and* usable by MakeMaker:
-$VERSION = '1.012.2';
+$VERSION = '1.012_5';
 
 ### Public Configuration Attributes
 $Config = {
@@ -171,10 +171,14 @@ sub _decode_B {
 # _decode_Q STRING
 #     Private: used by _decode_header() to decode "Q" encoding, which is
 #     almost, but not exactly, quoted-printable.  :-P
+#     Improvement by this module: sanity check on encoded sequence (>=1.012.3).
 sub _decode_Q {
     my $str = shift;
-    $str =~ s/_/\x20/g;					# RFC-1522, Q rule 2
-    $str =~ s/=([\da-fA-F]{2})/pack("C", hex($1))/ge;	# RFC-1522, Q rule 1
+    if ($str =~ /=(?![0-9a-fA-F][0-9a-fA-F])/) { #XXX:" " and "\t" are allowed
+	return undef;
+    }
+    $str =~ s/_/\x20/g;					# RFC 2047, Q rule 2
+    $str =~ s/=([0-9a-fA-F]{2})/pack("C", hex($1))/ge;	# RFC 2047, Q rule 1
     $str;
 }
 
@@ -220,6 +224,10 @@ B<**>
 However, adjacent encoded-words with same charset will be concatenated
 to handle multibyte sequences safely.
 
+B<**>
+Language information defined by RFC2231, section 5 will be additonal
+third element, if any.
+
 B<*>
 Whitespaces surrounding unencoded data will not be stripped so that
 compatibility with L<MIME::Words> will be ensured.
@@ -239,7 +247,7 @@ get I<something> back when decoding headers).
 $@ will be false if no error was detected.
 
 B<*>
-Malformed base64 encoded-words will be kept encoded.
+Malformed encoded-words will be kept encoded.
 In this case $@ will be set.
 
 Any arguments past the ENCODED string are taken to define a hash of options.
@@ -345,6 +353,7 @@ sub decode_mimewords {
 		next;
 	    }
 
+	  { local $@;
 	    if (scalar(@tokens) and
 		lc($charset || "") eq lc($tokens[-1]->[1] || "") and
 		resolve_alias($charset) and
@@ -359,6 +368,7 @@ sub decode_mimewords {
 		push @tokens, [$dec];
 	    }
 	    $spc = $tspc;
+	  }
             next;
         }
 
@@ -393,6 +403,7 @@ sub decode_mimewords {
 
     # Detect 7-bit charset
     if ($Params{Detect7bit} ne "NO") {
+	local $@;
 	foreach my $t (@tokens) {
 	    unless ($t->[0] =~ $UNSAFE or $t->[1]) {
 		my $charset = MIME::Charset::_detect_7bit_charset($t->[0]);
@@ -403,9 +414,13 @@ sub decode_mimewords {
 	}
     }
 
-    return (wantarray ? @tokens : join('',map {
-	&_convert($_->[0], $_->[1], $cset, $Params{Mapping})
-	} @tokens));
+    if (wantarray) {
+	@tokens;
+    } else {
+	join('', map {
+	    &_convert($_->[0], $_->[1], $cset, $Params{Mapping})
+	} @tokens);
+    }
 }
 
 #------------------------------
@@ -603,7 +618,8 @@ If empty string C<""> is specified, encoded-words exceeding line length
 
 B<Note>:
 B<*>
-Though RFC 2822 states that the lines are delimited by CRLF (C<"\r\n">), 
+Though RFC 5322 (formerly RFC 2822) states that the lines in
+Internet messages are delimited by CRLF (C<"\r\n">), 
 this module chose LF (C<"\n">) as a default to keep backward compatibility.
 When you use the default, you might need converting newlines
 before encoded headers are thrown into session.
@@ -624,6 +640,7 @@ B<**>
 
 Maximum line length excluding newline.
 The default is 76.
+Negative value means unlimited line length (as of release 1.012.3).
 
 =item Minimal
 B<**>
@@ -636,7 +653,7 @@ encoded-words exceeding line length will be split based only on their
 lengths.
 Default is C<"YES"> by which minimal portions of text are encoded.
 If C<"DISPNAME"> is specified, portions including special characters
-described in RFC5322 (former RFC2822, RFC822) address specification
+described in RFC5322 (formerly RFC2822, RFC822) address specification
 (section 3.4) are also encoded.
 This is useful for encoding display-name of address fields.
 
@@ -785,8 +802,10 @@ sub encode_mimewords  {
 				Replacement => $Params{Replacement},
 				Encoding => $Params{Encoding});
 	# Resolve 'S' encoding based on global length. See (*).
-	$enc = 'S' if $Params{Encoding} =~ /[AS]/ and
-		      $enc and $obj->header_encoding eq 'S';
+	$enc = 'S'
+	    if defined $enc and
+	       ($Params{Encoding} eq 'S' or
+		$Params{Encoding} eq 'A' and $obj->header_encoding eq 'S');
 
 	# pure ASCII
 	if ($cset eq "US-ASCII" and !$enc and $s =~ /$UNSAFEASCII/) {
@@ -854,14 +873,14 @@ sub encode_mimewords  {
 	} elsif ($Params{Encoding} eq 'A' and ! $b and $q) {
 	    $enc = 'Q';
 	# Otherwise, assuming 'Q', when characters to be encoded are more than
-	# 6th of total, 'B' will win.
+	# 6th of total (plus a little fraction), 'B' will win.
 	# Note: This might give 'Q' so great advantage...
 	} else {
 	    my @no_enc = grep { ! $_->[1] } @triplets;
 	    my $total = length join('', map { $_->[0] } (@no_enc, @s_enc));
 	    my $q = scalar(() = join('', map { $_->[0] } @s_enc) =~
 			   m{[^- !*+/0-9A-Za-z]}g);
-	    if ($total < $q * 6) {
+	    if ($total + 8 < $q * 6) {
 		$enc = 'B';
 	    } else {
 		$enc = 'Q';
@@ -878,27 +897,34 @@ sub encode_mimewords  {
     }
 
     # Split long ``words''.
-    my @splitted;
-    my $restlen = $firstlinelen;
-    foreach (@triplets) {
+    my @splitwords;
+    my $restlen;
+    if ($Params{MaxLineLen} < 0) {
+      @splitwords = @triplets;
+    } else {
+      $restlen = $firstlinelen;
+      foreach (@triplets) {
 	my ($s, $enc, $csetobj) = @$_;
 
-	push @splitted, &_split($s, $enc, $csetobj, $restlen, $maxrestlen);
-	my ($last, $lastenc, $lastcsetobj) = @{$splitted[-1]};
+	my @s = &_split($s, $enc, $csetobj, $restlen, $maxrestlen);
+	push @splitwords, @s;
+	my ($last, $lastenc, $lastcsetobj) = @{$s[-1]};
 	my $lastlen;
 	if ($lastenc) {
 	    $lastlen = $lastcsetobj->encoded_header_len($last, $lastenc);
 	} else {
 	    $lastlen = length($last);
 	}
-	$restlen -= $lastlen; # FIXME: Sometimes estimated longer
+	$restlen = $maxrestlen if scalar @s > 1; # has split; new line(s) fed
+	$restlen -= $lastlen;
 	$restlen = $maxrestlen if $restlen <= 1;
+      }
     }
 
     # Do encoding.
     my @lines;
     $restlen = $firstlinelen;
-    foreach (@splitted) {
+    foreach (@splitwords) {
 	my ($str, $encoding, $charsetobj) = @$_;
 	next unless length($str);
 
@@ -913,6 +939,8 @@ sub encode_mimewords  {
 		   $s =~ /^[\r\n\t ]/)? '': ' ';
 	if (!scalar(@lines)) {
 	    push @lines, $s;
+	} elsif ($Params{MaxLineLen} < 0) {
+	    $lines[-1] .= $spc.$s;
 	} elsif (length($lines[-1].$spc.$s) <= $restlen) {
 	    $lines[-1] .= $spc.$s;
 	} else {
@@ -952,10 +980,10 @@ sub _split {
 	return ([$str, $encoding, $charset]);
     }
 
-    my (@splitted, $ustr, $first);
+    my (@splitwords, $ustr, $first);
     while (length($str)) {
 	if ($charset->encoded_header_len($str, $encoding) <= $restlen) {
-	    push @splitted, [$str, $encoding, $charset];
+	    push @splitwords, [$str, $encoding, $charset];
 	    last;
 	}
 	$ustr = $str;
@@ -970,10 +998,10 @@ sub _split {
 	    ($first, $str) = &_clip_unsafe($ustr, $encoding, $charset,
 					   $maxrestlen);
 	}
-	push @splitted, [$first, $encoding, $charset];
+	push @splitwords, [$first, $encoding, $charset];
 	$restlen = $maxrestlen;
     }
-    return @splitted;
+    return @splitwords;
 }
 
 # _split_ascii RAW, ROOM_OF_FIRST_LINE, MAXRESTLEN
@@ -986,12 +1014,12 @@ sub _split_ascii {
     my $maxrestlen = shift;
     $restlen ||= $maxrestlen;
 
-    my @splitted;
+    my @splitwords;
     my $ascii = MIME::Charset->new("US-ASCII", Mapping => 'STANDARD');
     foreach my $line (split(/(?:[\t ]*[\r\n]+)+/, $s)) {
         my $spc = '';
 	foreach my $word (split(/([\t ]+)/, $line)) {
-	    next unless scalar(@splitted) or $word; # skip first garbage
+	    next unless scalar(@splitwords) or $word; # skip first garbage
 	    if ($word =~ /[\t ]/) {
 		$spc = $word;
 		next;
@@ -1000,33 +1028,33 @@ sub _split_ascii {
 	    my $cont = $spc.$word;
 	    my $elen = length($cont);
 	    next unless $elen;
-	    if (scalar(@splitted)) {
+	    if (scalar(@splitwords)) {
 		# Concatenate adjacent words so that encoded-word and
 		# unencoded text will adjoin with separating whitespace.
 		if ($elen <= $restlen) {
-		    $splitted[-1]->[0] .= $cont;
+		    $splitwords[-1]->[0] .= $cont;
 		    $restlen -= $elen;
 		} else {
-		    push @splitted, [$cont, undef, $ascii];
+		    push @splitwords, [$cont, undef, $ascii];
 		    $restlen = $maxrestlen - $elen;
 		}
 	    } else {
-		push @splitted, [$cont, undef, $ascii];
+		push @splitwords, [$cont, undef, $ascii];
 		$restlen -= $elen;
 	    }
 	    $spc = '';
 	}
 	if ($spc) {
-	    if (scalar(@splitted)) {
-		$splitted[-1]->[0] .= $spc;
+	    if (scalar(@splitwords)) {
+		$splitwords[-1]->[0] .= $spc;
 		$restlen -= length($spc);
 	    } else { # only WSPs
-		push @splitted, [$spc, undef, $ascii];
+		push @splitwords, [$spc, undef, $ascii];
 		$restlen = $maxrestlen - length($spc);
 	    }
 	}
     }
-    return @splitted;
+    return @splitwords;
 }
 
 # _clip_unsafe UNICODE, ENCODING, CHARSET_OBJECT, ROOM_OF_FIRST_LINE
@@ -1044,7 +1072,7 @@ sub _clip_unsafe {
     # Seek maximal division point.
     my ($shorter, $longer) = (0, length($ustr));
     while ($shorter < $longer) {
-	my $cur = int(($shorter + $longer + 1) / 2);
+	my $cur = ($shorter + $longer + 1) >> 1;
 	my $enc = substr($ustr, 0, $cur);
 	if (MIME::Charset::USE_ENCODE ne '') {
 	    $enc = $charset->undecode($enc);
